/*
 *  TOPPERS Software
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015 Cores Co., Ltd. Japan
 *
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 *
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 *
 *  @(#) $Id: fbs_string.c 164 2016-03-07 11:33:50Z coas-nagasima $
 */

#include <stdint.h>
#include <limits.h>
#include <stdarg.h>
#include "fbs_string.h"

/*
 *  l𕶎ɕϊ
 */
#define CONVERT_BUFLEN	((sizeof(uintptr_t) * CHAR_BIT + 2) / 3)
										/* uintptr_t^̐l̍ő啶 */
static ER
convert(ECN_FBS_ID fbs, ECN_FBS_SSIZE_T *pos, uintptr_t val, uint_t radix, const char *radchar,
			uint_t width, bool_t minus, bool_t padzero)
{
	char	buf[CONVERT_BUFLEN];
	uint_t	i, j;
	ER ret;

	i = 0U;
	do {
		buf[i++] = radchar[val % radix];
		val /= radix;
	} while (i < CONVERT_BUFLEN && val != 0);

	if (minus && width > 0) {
		width -= 1;
	}
	if (minus && padzero) {
		ret = _ecn_fbs_poke(fbs, *pos, '-');
		if(ret != E_OK)
			return ret;
		(*pos)++;
	}
	for (j = i; j < width; j++) {
		ret = _ecn_fbs_poke(fbs, *pos, padzero ? '0' : ' ');
		if(ret != E_OK)
			return ret;
		(*pos)++;
	}
	if (minus && !padzero) {
		ret = _ecn_fbs_poke(fbs, *pos, '-');
		if(ret != E_OK)
			return ret;
		(*pos)++;
	}
	while (i > 0U) {
		ret = _ecn_fbs_poke(fbs, *pos, buf[--i]);
		if(ret != E_OK)
			return ret;
		(*pos)++;
	}
	return E_OK;
}

/*
 *  񐮌`o
 */
static const char raddec[] = "0123456789";
static const char radhex[] = "0123456789abcdef";
static const char radHEX[] = "0123456789ABCDEF";

ECN_FBS_SSIZE_T
fbs_printf(ECN_FBS_ID fbs, ECN_FBS_SSIZE_T *pos, const char *format, ...)
{
	char		c;
	bool_t		lflag;
	uint_t		width;
	bool_t		padzero;
	intptr_t	val;
	const char	*str, *head;
	va_list	ap;
	ER ret;

	va_start(ap, format);

	while ((c = *format++) != '\0') {
		if (c != '%') {
			ret = _ecn_fbs_poke(fbs, *pos, c);
			if (ret != E_OK)
				return ret;
			(*pos)++;
			continue;
		}

		lflag = false;
		width = 0U;
		padzero = false;
		if ((c = *format++) == '0') {
			padzero = true;
			c = *format++;
		}
		while ('0' <= c && c <= '9') {
			width = width * 10U + c - '0';
			c = *format++;
		}
		if (c == 'l') {
			lflag = true;
			c = *format++;
		}
		switch (c) {
		case 'd':
			val = lflag ? (intptr_t) va_arg(ap, long_t)
						: (intptr_t) va_arg(ap, int_t);
			if (val >= 0) {
				ret = convert(fbs, pos, (uintptr_t) val, 10U, raddec,
										width, false, padzero);
				if (ret != E_OK)
					return ret;
			}
			else {
				ret = convert(fbs, pos, (uintptr_t)(-val), 10U, raddec,
										width, true, padzero);
				if (ret != E_OK)
					return ret;
			}
			break;
		case 'u':
			val = lflag ? (intptr_t) va_arg(ap, ulong_t)
						: (intptr_t) va_arg(ap, uint_t);
			ret = convert(fbs, pos, (uintptr_t) val, 10U, raddec, width, false, padzero);
			if (ret != E_OK)
				return ret;
			break;
		case 'x':
		case 'p':
			val = lflag ? (intptr_t) va_arg(ap, ulong_t)
						: (intptr_t) va_arg(ap, uint_t);
			ret = convert(fbs, pos, (uintptr_t) val, 16U, radhex, width, false, padzero);
			if (ret != E_OK)
				return ret;
			break;
		case 'X':
			val = (intptr_t)va_arg(ap, void *);
			ret = convert(fbs, pos, (uintptr_t) val, 16U, radHEX, width, false, padzero);
			if (ret != E_OK)
				return ret;
			break;
		case 'c':
			ret = _ecn_fbs_poke(fbs, *pos, (char)(intptr_t)va_arg(ap, int));
			if (ret != E_OK)
				return ret;
			(*pos)++;
			break;
		case 's':
			str = head = (const char *)va_arg(ap, const char *);
			while ((c = *str++) != '\0') {
				ret = _ecn_fbs_poke(fbs, *pos, c);
				if (ret != E_OK)
					return ret;
				(*pos)++;
			}
			break;
		case '%':
			ret = _ecn_fbs_poke(fbs, *pos, '%');
			if (ret != E_OK)
				return ret;
			(*pos)++;
			break;
		case '\0':
			format--;
			break;
		default:
			break;
		}
	}
	va_end(ap);

	return E_OK;
}
